@@ -15,13 +15,13 @@ except ImportError: #python 2.6
     from ordereddict import OrderedDict
 from collections import namedtuple
 import re
-
-if PY2:
-    import string
+import string
+import warnings
+from math import ceil
 
 dna_bases = re.compile(r'([ACTGNactgnYRWSKMDVHBXyrwskmdvhbx]+)')
 
-__version__ = '0.3.7.1'
+__version__ = '0.4.2'
 
 
 class FastaIndexingError(Exception):
@@ -55,10 +55,12 @@ class Sequence(object):
     start, end = coordinates of subsequence (optional)
     comp = boolean switch for complement property
     """
-    def __init__(self, name='', seq='', start=None, end=None, comp=False):
+    def __init__(self, name='', seq='', start=None, end=None, comp=False, one_based_attributes=True):
         self.name = name
         self.seq = seq
         self.start = start
+        if not one_based_attributes and start is not None:
+            self.start -= 1
         self.end = end
         self.comp = comp
         assert isinstance(name, string_types)
@@ -145,6 +147,28 @@ class Sequence(object):
         """
         return self[::-1]
 
+    @property
+    def orientation(self):
+        """ get the orientation forward=1, reverse=-1
+        >>> x = Sequence(name='chr1', seq='ATCGTA', start=1, end=6)
+        >>> x.orientation
+        1
+        >>> x.complement.orientation is None
+        True
+        >>> x[::-1].orientation is None
+        True
+        >>> x = -x
+        >>> x.orientation
+        -1
+        """
+        if self.start < self.end and not self.comp:
+            return 1
+        elif self.start > self.end and self.comp:
+            return -1
+        else:
+            return None
+
+
     @property
     def gc(self):
         """ Return the GC content of seq as a float
@@ -160,7 +184,7 @@ class Sequence(object):
         return (g + c) / len(self.seq)
 
 
-class IndexRecord(namedtuple('IndexRecord', ['rlen', 'offset', 'lenc', 'lenb'])):
+class IndexRecord(namedtuple('IndexRecord', ['rlen', 'offset', 'lenc', 'lenb', 'bend', 'prev_bend'])):
     __slots__ = ()
 
     def __getitem__(self, key):
@@ -176,7 +200,9 @@ class Faidx(object):
     """ A python implementation of samtools faidx FASTA indexing """
     def __init__(self, filename, default_seq=None, key_function=None,
                  as_raw=False, strict_bounds=False, read_ahead=None,
-                 mutable=False, split_char=None):
+                 mutable=False, split_char=None, filt_function=None,
+                 one_based_attributes=True, 
+                 sequence_always_upper=False):
         """
         filename: name of fasta file
         key_function: optional callback function which should return a unique
@@ -192,9 +218,12 @@ class Faidx(object):
             self.file = open(filename, 'rb')
         self.indexname = filename + '.fai'
         self.key_function = key_function if key_function else lambda rname: rname
+        self.filt_function = filt_function if filt_function else lambda x: True
         self.as_raw = as_raw
         self.default_seq = default_seq
         self.strict_bounds = strict_bounds
+        self.one_based_attributes = one_based_attributes
+        self.sequence_always_upper = sequence_always_upper
         self.index = OrderedDict()
         self.buffer = dict((('seq', None), ('name', None), ('start', None), ('end', None)))
         if not read_ahead or isinstance(read_ahead, int):
@@ -229,21 +258,27 @@ class Faidx(object):
     def read_fai(self, split_char):
         duplicate_ids = []
         with open(self.indexname) as index:
+            prev_bend = 0
             for line in index:
-                line = line.strip()
+                line = line.rstrip()
                 rname, rlen, offset, lenc, lenb = line.split('\t')
-                rname = self.key_function(rname).split(split_char)
-                for key in rname:
-                    if key in self.index and not split_char:
-                        raise ValueError('Duplicate key "%s"' % rname)
+                rlen, offset, lenc, lenb = map(int, (rlen, offset, lenc, lenb))
+                newlines = int(ceil(rlen / lenc) * (lenb - lenc))
+                bend = offset + newlines + rlen
+                rname = filter(self.filt_function, self.key_function(rname).split(split_char))
+                for i, key in enumerate(rname):  # mdshw5/pyfaidx/issues/64
+                    if key in self.index and split_char is None:
+                        if i == 0:
+                            raise ValueError('Duplicate key "%s"' % key)
+                        else:
+                            continue
+                    # eliminate duplicate keys if they result from split_char
                     elif key in self.index and split_char:
                         duplicate_ids.append(key)
                         continue
                     else:
-                        self.index[key] = IndexRecord(*map(int, (rlen,
-                                                                   offset,
-                                                                   lenc,
-                                                                   lenb)))
+                        self.index[key] = IndexRecord(rlen, offset, lenc, lenb, bend, prev_bend)
+                prev_bend = bend
             for dup in duplicate_ids:
                 self.index.pop(dup, None)
 
@@ -254,15 +289,15 @@ class Faidx(object):
                 raise FastaIndexingError("Line length of fasta"
                                          " file is not "
                                          "consistent! "
-                    "Inconsistent line found in >{0} at "
-                    "line {1:n}.".format(rname, bad_lines[0] + 1))
+                                         "Inconsistent line found in >{0} at "
+                                         "line {1:n}.".format(rname, bad_lines[0] + 1))
             elif len(bad_lines) == 1:  # check that the line is previous line
                 if bad_lines[0] + 1 != i:
                     raise FastaIndexingError("Line length of fasta"
                                              " file is not "
                                              "consistent! "
-                        "Inconsistent line found in >{0} at "
-                        "line {1:n}.".format(rname, bad_lines[0] + 1))
+                                             "Inconsistent line found in >{0} at "
+                                             "line {1:n}.".format(rname, bad_lines[0] + 1))
 
         with open(self.filename, 'r') as fastafile:
             with open(self.indexname, 'w') as indexfile:
@@ -358,19 +393,16 @@ class Faidx(object):
 
 
         # Calculate offset (https://github.com/samtools/htslib/blob/20238f354894775ed22156cdd077bc0d544fa933/faidx.c#L398)
-        newlines_total = int(i.rlen / i.lenc * (i.lenb - i.lenc))
-        newlines_before = int((start0 - 1) / i.lenc *
-                              (i.lenb - i.lenc)) if start0 > 0 else 0
+        newlines_before = int((start0 - 1) / i.lenc * (i.lenb - i.lenc)) if start0 > 0 else 0
         newlines_to_end = int(end / i.lenc * (i.lenb - i.lenc))
         newlines_inside = newlines_to_end - newlines_before
         seq_blen = newlines_inside + seq_len
         bstart = i.offset + newlines_before + start0
-        bend = i.offset + newlines_total + i.rlen
         self.file.seek(bstart)
 
-        if bstart + seq_blen > bend and not self.strict_bounds:
-            seq_blen = bend - bstart
-        elif bstart + seq_blen > bend and self.strict_bounds:
+        if bstart + seq_blen > i.bend and not self.strict_bounds:
+            seq_blen = i.bend - bstart
+        elif bstart + seq_blen > i.bend and self.strict_bounds:
             raise FetchError("Requested end coordinate {0:n} outside of {1}. "
                              "\n".format(end, rname))
         if seq_blen > 0:
@@ -392,12 +424,15 @@ class Faidx(object):
             seq = ''.join([seq, pad_len * self.default_seq])
         else:  # Return less than requested range
             end = start0 + len(seq)
-
+        
+        if self.sequence_always_upper:
+            seq = seq.upper()
+            
         if self.as_raw:
             return seq
         else:
             return Sequence(name=rname, start=int(start),
-                            end=int(end), seq=seq)
+                            end=int(end), seq=seq, one_based_attributes=self.one_based_attributes)
 
     def to_file(self, rname, start, end, seq):
         """ Write sequence in region from start-end, overwriting current
@@ -435,7 +470,6 @@ class FastaRecord(object):
     def __init__(self, name, fa):
         self.name = name
         self._fa = fa
-        self._len = self._fa.faidx.index[self.name].rlen
 
     def __getitem__(self, n):
         """Return sequence from region [start, end)
@@ -478,25 +512,34 @@ class FastaRecord(object):
         return 'FastaRecord("%s")' % (self.name)
 
     def __len__(self):
-        return self._len
+        return self._fa.faidx.index[self.name].rlen
 
     def __str__(self):
         return str(self[:])
 
+    @property
+    def variant_sites(self):
+        if isinstance(self._fa, FastaVariant):
+            pos = []
+            var = self._fa.vcf.fetch(self.name, 0, len(self))
+            for site in var:
+                if site.is_snp:
+                    sample = site.genotype(self._fa.sample)
+                    if sample.gt_type in self._fa.gt_type and eval(self._fa.filter):
+                        pos.append(site.POS)
+            return tuple(pos)
+        else:
+            raise NotImplementedError("variant_sites() only valid for FastaVariant.")
+
     @property
     def long_name(self):
         """ Read the actual defline from self._fa.faidx mdshw5/pyfaidx#54 """
-        seqnames = tuple(self._fa.keys())
-        seqname_i = seqnames.index(self.name)
-        if seqname_i > 0:
-            seqname_p = seqnames[seqname_i - 1]  # previous seqname
-            _, internals = self._fa.faidx.from_file(seqname_p, 1, 1, internals=True)
-            defline_start = internals['bend'] + 1
-        elif seqname_i == 0:
-            defline_start = 0
-        defline_end = self._fa.faidx.index[self.name].offset
-        self._fa.faidx.file.seek(defline_start)
-        return self._fa.faidx.file.read(defline_end - defline_start).decode()[1:-1]
+        index_record = self._fa.faidx.index[self.name]
+        prev_bend = index_record.prev_bend
+        newline_len = index_record.lenb - index_record.lenc
+        defline_end = index_record.offset
+        self._fa.faidx.file.seek(prev_bend)
+        return self._fa.faidx.file.read(defline_end - prev_bend).decode()[1:-1]
 
 class MutableFastaRecord(FastaRecord):
     def __init__(self, name, fa):
@@ -530,7 +573,10 @@ class MutableFastaRecord(FastaRecord):
 
 
 class Fasta(object):
-    def __init__(self, filename, default_seq=None, key_function=None, as_raw=False, strict_bounds=False, read_ahead=None, mutable=False, split_char=None):
+    def __init__(self, filename, default_seq=None, key_function=None, as_raw=False, 
+                 strict_bounds=False, read_ahead=None, mutable=False, split_char=None, 
+                 filt_function=None, one_based_attributes=True, 
+                 sequence_always_upper=False):
         """
         An object that provides a pygr compatible interface.
         filename: name of fasta file
@@ -539,7 +585,9 @@ class Fasta(object):
         self.mutable = mutable
         self.faidx = Faidx(filename, key_function=key_function, as_raw=as_raw,
                            default_seq=default_seq, strict_bounds=strict_bounds,
-                           read_ahead=read_ahead, mutable=mutable, split_char=split_char)
+                           read_ahead=read_ahead, mutable=mutable, split_char=split_char,
+                           filt_function=filt_function, one_based_attributes=one_based_attributes,
+                           sequence_always_upper=sequence_always_upper)
         self.keys = self.faidx.index.keys
         if not self.mutable:
             self.records = dict([(rname, FastaRecord(rname, self)) for rname in self.keys()])
@@ -584,6 +632,78 @@ class Fasta(object):
         self.faidx.__exit__(*args)
 
 
+class FastaVariant(Fasta):
+    """ Return consensus sequence from FASTA and VCF inputs
+    """
+    expr = set(('>', '<', '=', '!'))
+    def __init__(self, filename, vcf_file, sample=None, het=True, hom=True, call_filter=None, **kwargs):
+        try:
+            import pysam
+        except ImportError:
+            raise ImportError("pysam must be installed for FastaVariant.")
+        try:
+            import vcf
+        except ImportError:
+            raise ImportError("PyVCF must be installed for FastaVariant.")
+        if call_filter is not None:
+            try:
+                key, expr, value = call_filter.split() # 'GQ > 30'
+            except IndexError:
+                raise ValueError("call_filter must be a string in the format 'XX <>!= NN'")
+            assert all([x in self.expr for x in list(expr)])
+            assert all([x in string.ascii_uppercase for x in list(key)])
+            assert all([x in string.printable for x in list(value)])
+            self.filter = "sample['{key}'] {expr} {value}".format(**locals())
+        else:
+            self.filter = 'True'
+        self.vcf = vcf.Reader(filename=vcf_file)
+        if sample is not None:
+            self.sample = sample
+        else:
+            self.sample = self.vcf.samples[0]
+            if len(self.vcf.samples) > 1 and sample is None:
+                warnings.warn("Using sample {0} genotypes.".format(self.sample), RuntimeWarning)
+        if het and hom:
+            self.gt_type = set((1, 2))
+        elif het:
+            self.gt_type = set((1,))
+        elif hom:
+            self.gt_type = set((2,))
+        else:
+            self.gt_type = set()
+        super(FastaVariant, self).__init__(filename, **kwargs)
+
+    def __repr__(self):
+        return 'FastaVariant("%s", "%s", gt="%s")' % (self.filename, self.vcf.filename, str(self.gt_type))
+
+    def get_seq(self, name, start, end):
+        """Return a sequence by record name and interval [start, end).
+        Replace positions with polymorphism with variant.
+        Coordinates are 0-based, end-exclusive.
+        """
+        seq = self.faidx.fetch(name, start, end)
+        if self.faidx.as_raw:
+            seq_mut = list(seq)
+            del seq
+        else:
+            seq_mut = list(seq.seq)
+            del seq.seq
+        var = self.vcf.fetch(name, start, end)
+        for record in var:
+            if record.is_snp:  # skip indels
+                sample = record.genotype(self.sample)
+                if sample.gt_type in self.gt_type and eval(self.filter):
+                    alt = record.ALT[0]
+                    i = (record.POS - 1) - (start - 1)
+                    seq_mut[i:i + len(alt)] = str(alt)
+        # slice the list in case we added an MNP in last position
+        if self.faidx.as_raw:
+            return ''.join(seq_mut[:end - start + 1])
+        else:
+            seq.seq = ''.join(seq_mut[:end - start + 1])
+            return seq
+
+
 def wrap_sequence(n, sequence, fillvalue=''):
     args = [iter(sequence)] * n
     for line in zip_longest(fillvalue=fillvalue, *args):
